package Instance;

import java.util.ArrayList;

public class IndexedInstance {
    public double[] tupleWeight;
    private int nodes;
    private int edges;
    private int tuples;
    private int[] domainSize;
    private int[] edgeFromNodeId;
    private int[] edgeToNodeId;
    private int[] scopeIdToFirstTupleId;
    private int[] tupleIdToScopeId;
    //private int[] tupleIdToLabel1;
    //private int[] tupleIdToLabel2;
    private ArrayList<int[]> incidentEdgesToNodeId;

    public IndexedInstance(Instance instance){
        tuples = instance.totalTuples;
        tupleWeight = new double[instance.totalTuples+1];
        nodes = Math.max(instance.nodes, instance.nodesCheck);
        edges = instance.edges;
        domainSize = new int[nodes+1];
        for (int i = 1; i <= nodes; i++) {
            domainSize[i] = instance.domainSize[i-1];
        }
        edgeFromNodeId = new int[edges+1];
        edgeToNodeId = new int[edges+1];
        scopeIdToFirstTupleId = new int[nodes+edges+1];
        tupleIdToScopeId = new int[instance.totalTuples+1];
        //tupleIdToLabel1 = new int[instance.totalTuples+1];
        //tupleIdToLabel2 = new int[instance.totalTuples+1];
        incidentEdgesToNodeId = new ArrayList<int[]>();
        for (int i = 0; i < nodes; i++) {
            incidentEdgesToNodeId.add(new int[instance.edgesIncidentToNode[i]]);
        }
        int[] edgesEnteredForNodeId = new int[nodes+1];

        int factor = 0;
        int freeEdge = 1;
        int freeTuple = 1;
        /*for (int factor = 0; factor < instance.factorValues.size(); factor++) {
            long[][] values = instance.factorValues.get(factor);*/
        //}
        for (long[][] values:instance.factorValues) {
            if(instance.factorArity[factor] == 1){
                // node
                int nodeId = instance.factorScope[factor][0]+1;
                int scope = nodeIdToScopeId(nodeId);
                scopeIdToFirstTupleId[scope] = freeTuple;
                // process tuples for this node
                for (int i = 0; i < domainSize[nodeId]; i++) {
                    tupleWeight[freeTuple] = values[i][0];
                    //tupleIdToLabel1[freeTuple] = i;
                    tupleIdToScopeId[freeTuple] = scope;
                    //tupleIdToLabel2[freeTuple] = -1;// null value

                    freeTuple++;
                }


            } else if(instance.factorArity[factor] == 2){
                // edge
                int edgeId = freeEdge++;
                int scope = edgeIdToScopeId(edgeId);
                int from = instance.factorScope[factor][0]+1;
                int to = instance.factorScope[factor][1]+1;
                edgeFromNodeId[edgeId] = from;
                edgeToNodeId[edgeId] = to;

                incidentEdgesToNodeId.get(from-1)[edgesEnteredForNodeId[from]++] = edgeId; // positive if node is from
                incidentEdgesToNodeId.get(to-1)[edgesEnteredForNodeId[to]++] = -edgeId; // negative for opposite direction

                scopeIdToFirstTupleId[scope] = freeTuple;
                // process tuples for this edge
                for (int i = 0; i < domainSize[from]; i++) {
                    for (int j = 0; j < domainSize[to]; j++) {
                        tupleWeight[freeTuple] = values[i][j];
                        //tupleIdToLabel1[freeTuple] = i;
                        //tupleIdToLabel2[freeTuple] = j;
                        tupleIdToScopeId[freeTuple] = scope;
                        //System.out.println("Edge from " + from + " to " + to + " labels (" + i + ", "+ j + "), weight: " + values[i][j]);

                        freeTuple++;
                    }
                }
            }

            factor++;
        }
    }

    public int tuples(){
        return tuples;
    }

    public int[] incidentEdgesToNodeId(int nodeId){
        return incidentEdgesToNodeId.get(nodeId-1);
    }

    public int edgeIdFrom(int edgeId){
        if(edgeId < 0){
            return edgeIdTo(-edgeId);
        }
        return edgeFromNodeId[edgeId];
    }

    public int edgeIdTo(int edgeId) {
        if(edgeId < 0){
            return edgeIdFrom(-edgeId);
        }
        return edgeToNodeId[edgeId];
    }

    public int nodes(){
        return nodes;
    }

    public int edges(){
        return edges;
    }

    public int scopes(){
        return nodes+edges;
    }

    public int tupleIdToScopeId(int tupleId){
        return tupleIdToScopeId[tupleId];
    }

    /*public int tupleIdToLabel1(int tupleId){
        return tupleIdToLabel1[tupleId];
    }*/

    /*public int tupleIdToLabel2(int tupleId){
        return tupleIdToLabel2[tupleId];
    }*/

    public int nodeIdLabelToTupleId(int nodeId, int label){
        int scope = nodeIdToScopeId(nodeId);
        int firstTuple = scopeIdToFirstTupleId[scope];
        return firstTuple + label;
    }

    public int edgeIdLabel1Label2ToTupleId(int edgeId, int label1, int label2){
        if(edgeId < 0){
            return edgeIdLabel1Label2ToTupleId(-edgeId, label2, label1);
        }
        int scope = edgeIdToScopeId(edgeId);
        int firstTuple = scopeIdToFirstTupleId[scope];
        return firstTuple + label1*domainSize(edgeToNodeId[edgeId])+label2;
    }

    public int domainSize(int nodeId){
        return domainSize[nodeId];
    }

    public int nodeIdToScopeId(int nodeId){
        return nodeId+edges;
    }

    public int edgeIdToScopeId(int edgeId){
        return Math.abs(edgeId);
    }

    public int scopeIdToNodeId(int scopeId){
        if(arity(scopeId) == 2){
            System.out.println("ERR: Assuming edge scope is node scope.");
        }
        return scopeId-edges;
    }

    public int scopeIdToEdgeId(int scopeId){
        if(arity(scopeId) == 1){
            System.out.println("ERR: Assuming node scope is edge scope.");
        }
        return scopeId;
    }

    public int arity(int scopeId){
        if(scopeId <= edges){
            return 2;
        } else {
            return 1;
        }
    }

    public int maxDomainSize() {
        int result = 0;
        for (int nodeId = 1; nodeId <= nodes(); nodeId++) {
            result = Math.max(result,domainSize(nodeId));
        }
        return result;
    }

    public double averageDegree() {
        double result = 0;
        for (int nodeId = 1; nodeId <= nodes(); nodeId++) {
            result += incidentEdgesToNodeId(nodeId).length;
        }
        result = result/nodes();
        return result;
    }
}
