package Instance;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.LinkedList;

public class Instance {

    public int nodes;
    private int maxDomainSize;
    private int numFactors;
    private long initialUB;
    public int[] domainSize;
    public LinkedList<long[][]> factorValues;
    public int[] factorArity;
    public int[][] factorScope;

    public int nodesCheck;
    public int edges;


    public int totalTuples;

    public int[] edgesIncidentToNode;


    public Instance(String filename) throws IOException, IllegalArgumentException {
        BufferedReader br = new BufferedReader(new FileReader(filename));
        try {
            // read first line
            String line = br.readLine();
            if(line.length() == 0) line = br.readLine();
            String[] split = line.trim().split("\\s++");
            nodes = Integer.parseInt(split[1]);
            if(nodes==0) throw new IllegalArgumentException("trivial instance");
            maxDomainSize = Integer.parseInt(split[2]);
            numFactors = Integer.parseInt(split[3]);
            initialUB = Long.parseLong(split[4]);

            // read second line - domain sizes
            domainSize = new int[nodes];
            line = br.readLine();
            split = line.trim().split("\\s++");
            for (int i = 0; i < nodes; i++) {
                domainSize[i] = Integer.parseInt(split[i]);
            }

            totalTuples = 0;
            edgesIncidentToNode = new int[nodes];

            factorValues = new LinkedList<>();
            factorArity = new int[numFactors];
            factorScope = new int[numFactors][2];
            for (int factor = 0; factor < numFactors; factor++) {
                line = br.readLine();
                split = line.trim().split("\\s++");
                int arity = Integer.parseInt(split[0]);
                factorArity[factor] = arity;
                if(arity == 1){
                    int var1 = Integer.parseInt(split[1]);
                    factorScope[factor][0] = var1;
                    long defaultCost = -Long.parseLong(split[2]);  // NEGATIVE because we minimize
                    long[][] values = new long[domainSize[var1]][1];
                    for (int i = 0; i < domainSize[var1]; i++) {
                        values[i][0] = defaultCost;
                    }
                    int numTuples = Integer.parseInt(split[3]);
                    for (int i = 0; i < numTuples; i++) {
                        line = br.readLine();
                        split = line.trim().split("\\s++");
                        int label1 = Integer.parseInt(split[0]);
                        long value = -Long.parseLong(split[1]); // NEGATIVE because we minimize
                        values[label1][0] = value;
                    }
                    factorValues.addLast(values);
                    totalTuples += domainSize[var1];
                    nodesCheck++;

                } else if (arity == 2){
                    int var1 = Integer.parseInt(split[1]);
                    int var2 = Integer.parseInt(split[2]);
                    factorScope[factor][0] = var1;
                    factorScope[factor][1] = var2;
                    edgesIncidentToNode[var1]++;
                    edgesIncidentToNode[var2]++;
                    long defaultCost = -Long.parseLong(split[3]);  // NEGATIVE because we minimize
                    long[][] values = new long[domainSize[var1]][domainSize[var2]];
                    for (int i = 0; i < domainSize[var1]; i++) {
                        for (int j = 0; j < domainSize[var2]; j++) {
                            values[i][j] = defaultCost;
                        }
                    }
                    int numTuples = Integer.parseInt(split[4]);
                    for (int i = 0; i < numTuples; i++) {
                        line = br.readLine();
                        split = line.trim().split("\\s++");
                        int label1 = Integer.parseInt(split[0]);
                        int label2 = Integer.parseInt(split[1]);
                        long value = -Long.parseLong(split[2]); // NEGATIVE because we minimize
                        values[label1][label2] = value;
                    }
                    factorValues.addLast(values);
                    totalTuples += domainSize[var1]*domainSize[var2];
                    edges++;

                } else if (arity == 0) {
                    long constantCost = -Long.parseLong(split[1]);  // NEGATIVE because we minimize
                    // add this constant cost to any factor (e.g. first)
                    long[][] vals = factorValues.getFirst();
                    for (int i = 0; i < vals.length; i++) {
                        for (int j = 0; j < vals[i].length; j++) {
                            vals[i][j] += constantCost;
                        }
                    }
                } else {
                    throw new IllegalArgumentException("ERR: Factor with arity >= 3, line: " + line);
                }
            }
        } finally {
            br.close();
        }
    }

}


