import Instance.Instance;
import Instance.IndexedInstance;
import Propagators.ArcConsistencyPropagator;
import Propagators.CycleConsistencyPropagator;
import Propagators.PropagatorInterface;
import Propagators.SingletonArcConsistencyPropagator;
import Toulbar2API.FileProcessor;
import Utils.CycleGenerator;

import java.io.*;
import java.util.Arrays;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

public class Main {

    public static void main(String[] args) throws IOException, InterruptedException {

        File[] inputFiles = createSortedInputFiles("/home/t440p/MyFolder/cost-function-library-master", true);

        System.out.println("Num input files: " + inputFiles.length);

        String outputFilename = "/home/t440p/MyFolder/results-textfiles/output_CFL_" + System.nanoTime() + ".txt";
        File outputFile = new File(outputFilename);
        BufferedWriter writer = new BufferedWriter(new FileWriter(outputFile));

        for (int j = 0; j < inputFiles.length; j++) {
            System.out.println("Processing instance " + j + " - " + inputFiles[j].getAbsolutePath());

            processInstance(inputFiles[j].getAbsolutePath(),writer,"/home/t440p/MyFolder/toulbar2-Released","/home/t440p/MyFolder/results-from-toulbar/",j);

        }
        writer.close();
    }

    private static void processInstance(String filename, BufferedWriter writer, String toulbar2Released, String targetDirectory, int instanceID) throws IOException,
            InterruptedException {

        try{
            Instance instance = new Instance(filename);
        } catch (IllegalArgumentException ex){
            System.out.println("Skipping instance " + filename);
            System.out.println(ex.getMessage());
            return;
        }

        System.out.println("EDAC ...");
        // first process EDAC, store output somewhere suitable and store instance
        String command = toulbar2Released + " " + filename + " -k=4 -bt=0 -z=2";
        long start = System.nanoTime();
        String EDACbound = FileProcessor.runCommandAndGetBound(command);
        long end = System.nanoTime();
        double EDACtime = ((end-start)*1.0)/1e9;
        //FileProcessor.renameFile("problem.wcsp",targetDirectory+instanceID+"_EDAC.wcsp");

        System.out.println("VAC ...");
        // then process VAC, store output somewhere suitable and store instance
        command = toulbar2Released + " " + filename + " -A -bt=0 -z=2";
        start = System.nanoTime();
        String VACbound = FileProcessor.runCommandAndGetBound(command);
        end = System.nanoTime();
        long VACtimeExact=end-start;
        double VACtime = ((end-start)*1.0)/1e9;
        FileProcessor.renameFile("problem.wcsp",targetDirectory+"instance_VAC.wcsp");

        // load VAC instance, put it to SAC and CC
        Instance instance;
        try{
           instance = new Instance(targetDirectory+"instance_VAC.wcsp");
        } catch (IllegalArgumentException ex){
            System.out.println("Skipping instance " + filename);
            System.out.println(ex.getMessage());
            return;
        }
        IndexedInstance indInstance = new IndexedInstance(instance);

        long timeLimitPerInstanceInSeconds = 3600;

        Optimizer opt = new Optimizer(indInstance);
        long timeLimit;
        PropagatorInterface prop;

        System.out.println("SAC ...");
        // run SAC
        start = System.nanoTime();
        timeLimit = System.nanoTime()+timeLimitPerInstanceInSeconds*1000000000L-VACtimeExact;
        prop = new SingletonArcConsistencyPropagator(indInstance);
        double SACbound = opt.getBound(prop,timeLimit);
        end = System.nanoTime();
        double SACtime = (end-start)/1e9+VACtime;

        System.out.println("CC ...");
        // run CC
        start = System.nanoTime();
        timeLimit = System.nanoTime()+timeLimitPerInstanceInSeconds*1000000000L-VACtimeExact;
        prop = new CycleConsistencyPropagator(indInstance);
        double CCbound = opt.getBound(prop,timeLimit);
        end = System.nanoTime();
        double CCtime = (end-start)/1e9+VACtime;


        System.out.println("pseudotriangles ...");
        // run pseudotriangles
        command = toulbar2Released + " " + filename + " -t=8000 -bt=0 -z=2";
        start = System.nanoTime();
        String pseudoTrianglesbound = FileProcessor.runCommandAndGetBound(command);
        end = System.nanoTime();
        double pseudoTrianglestime = ((end-start)*1.0)/1e9;
        //FileProcessor.renameFile("problem.wcsp",targetDirectory+instanceID+"_pseudotr.wcsp");

        // store results in a text file - EDAC, VAC, SAC, CC, pseudotr
        String reportBounds = ""+EDACbound+";"+VACbound+";"+SACbound+";"+CCbound+";"+pseudoTrianglesbound;
        String reportTime = EDACtime+";"+VACtime+";"+SACtime+";"+CCtime+";"+pseudoTrianglestime;
        String report = ""+instanceID+";"+filename+";"+reportBounds+";"+reportTime;
        System.out.println(filename+" was processed: ");
        System.out.println(report);
        writer.write(report);
        writer.newLine();
        writer.flush();

    }

    private static void listf(String directoryName, List<File> files) {
        File directory = new File(directoryName);

        // Get all files from a directory.
        File[] fList = directory.listFiles();
        if(fList != null)
            for (File file : fList) {
                if (file.isFile()) {
                    files.add(file);
                } else if (file.isDirectory()) {
                    listf(file.getAbsolutePath(), files);
                }
            }
    }

    private static File[] createSortedInputFiles(String directory, boolean sort){
        List<File> list = new LinkedList<>();
        listf(directory,list);

        List<File> inputFiles = new LinkedList<>();
        for (File file: list) {
            if(file.getAbsolutePath().endsWith("wcsp")){
                inputFiles.add(file);
            }
        }

        File[] sortedInputFiles = new File[inputFiles.size()];
        int i = 0;
        for (File file: inputFiles) {
            sortedInputFiles[i++] = file;
        }
        if(sort){
        Arrays.sort(sortedInputFiles, new Comparator<File>() {
            public int compare(File f1, File f2)
            {
                return Math.round(Math.signum(f1.length()-f2.length()));
            }
        });}
        return sortedInputFiles;
    }
}
