package Utils;

import Instance.IndexedInstance;

public class ArcConsistencyStruct {
    private IndexedInstance indexedInstance;
    private int[] queue;
    private boolean[] isInQueuePositive;
    private boolean[] isInQueueNegative;

    private int elemsInQueue;
    private int freeElemInQueue;
    private int topOfQueue;

    public ArcConsistencyStruct(IndexedInstance indexedInstance){
        this.indexedInstance = indexedInstance;

        queue = new int[2*indexedInstance.edges()];
        isInQueueNegative = new boolean[indexedInstance.edges()+1];
        isInQueuePositive = new boolean[indexedInstance.edges()+1];

        elemsInQueue = 0;
        freeElemInQueue = 0;
        topOfQueue = -1;
    }

    public void fillQueue(){
        elemsInQueue = 2*indexedInstance.edges();
        for (int i = 0; i < indexedInstance.edges(); i++) {
            queue[i] = i+1;
            isInQueuePositive[i+1] = true;
        }
        for (int i = 0; i < indexedInstance.edges(); i++) {
            queue[i+indexedInstance.edges()] = -(i+1);
            isInQueueNegative[i+1] = true;
        }
        topOfQueue = 0;
        freeElemInQueue = 0;
    }

    public int getElement(){
        if(isEmpty()){
            System.out.println("ERR: Querying empty queue");
            return 0;
        }

        int result = queue[topOfQueue];
        if(result>0){
            isInQueuePositive[result] = false;
        } else {
            isInQueueNegative[-result] = false;
        }

        topOfQueue++;
        if(topOfQueue>=queue.length){
            topOfQueue = 0;
        }

        elemsInQueue--;

        return result;
    }

    public void addElement(int signedEdgeId){
        if(signedEdgeId > 0 && !isInQueuePositive[signedEdgeId]){
            isInQueuePositive[signedEdgeId] = true;
            queue[freeElemInQueue] = signedEdgeId;
            freeElemInQueue++;
            elemsInQueue++;
            if(topOfQueue == -1){
                topOfQueue = 0;
            }
        } else if (signedEdgeId < 0 && !isInQueueNegative[-signedEdgeId]){
            isInQueueNegative[-signedEdgeId] = true;
            queue[freeElemInQueue] = signedEdgeId;
            freeElemInQueue++;
            elemsInQueue++;
            if(topOfQueue == -1){
                topOfQueue = 0;
            }
        }
        if(freeElemInQueue >= queue.length){
            freeElemInQueue = 0;
        }
    }

    public boolean isEmpty(){
        return elemsInQueue==0;
    }
}
