package Utils;

import Instance.IndexedInstance;

import java.util.LinkedList;
import java.util.Queue;

public class CycleGenerator {

    public static LinkedList<int[]> generateFundamentalCycles(IndexedInstance indexedInstance){
        int[] edgeToParent = new int[indexedInstance.nodes()+1];
        int[] treeParent = new int[indexedInstance.nodes()+1];
        // treeParent == 0 for nodes not in tree, treeParent != 0 for nodes in tree, treeParent == -1 for root
        treeParent[1] = -1;

        LinkedList<int[]> fundamentalCycles = new LinkedList<>();
        exploreSpanningTreeAndFundamentalCycles(1,indexedInstance,fundamentalCycles, treeParent, edgeToParent);
        return fundamentalCycles;
    }

    private static void exploreSpanningTreeAndFundamentalCycles(int currentNode, IndexedInstance indexedInstance,
                                                               LinkedList<int[]> fundamentalCycles, int[] treeParent,
                                                                int[] edgeToParent){
        Queue<Integer> queue = new LinkedList<>();
        queue.add(currentNode);
        boolean[]processed = new boolean[indexedInstance.nodes()+1];
        while(!queue.isEmpty()){
            currentNode = queue.poll();
            if(processed[currentNode]){
                continue;
            }
            for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(currentNode).length; i++) {
                int edgeId = indexedInstance.incidentEdgesToNodeId(currentNode)[i];
                int nodeIdTo = indexedInstance.edgeIdTo(edgeId);
                if(treeParent[nodeIdTo] != 0){
                    // add fundamental cycle
                    addFundamentalCycle(currentNode,nodeIdTo,edgeId,fundamentalCycles,treeParent,edgeToParent);
                } else {
                    treeParent[nodeIdTo] = currentNode;
                    edgeToParent[nodeIdTo] = -edgeId;
                    //exploreSpanningTreeAndFundamentalCycles(nodeIdTo,indexedInstance,fundamentalCycles,treeParent,edgeToParent);
                    queue.add(nodeIdTo);
                }
            }
            processed[currentNode] = true;
        }

    }

    /*private static void exploreSpanningTreeAndFundamentalCycles(int currentNode, IndexedInstance indexedInstance,
                                                                LinkedList<int[]> fundamentalCycles, int[] treeParent,
                                                                int[] edgeToParent){
            for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(currentNode).length; i++) {
                int edgeId = indexedInstance.incidentEdgesToNodeId(currentNode)[i];
                int nodeIdTo = indexedInstance.edgeIdTo(edgeId);
                if(treeParent[nodeIdTo] != 0){
                    // add fundamental cycle
                    addFundamentalCycle(currentNode,nodeIdTo,edgeId,fundamentalCycles,treeParent,edgeToParent);
                } else {
                    treeParent[nodeIdTo] = currentNode;
                    edgeToParent[nodeIdTo] = -edgeId;
                    exploreSpanningTreeAndFundamentalCycles(nodeIdTo,indexedInstance,fundamentalCycles,treeParent,edgeToParent);
                }
            }
    }*/

    private static void addFundamentalCycle(int nodeFrom, int nodeTo, int edgeId,
                                           LinkedList<int[]> fundamentalCycles, int[] treeParent, int[] edgeToParent){
        // find common ancestor of nodeTo,nodeFrom, denote as A
        int commonAncestor = findCommonAncestor(nodeTo,nodeFrom,treeParent);
        // find edges from nodeFrom to A, find edges from nodeTo to A, merge with edgeId, add as new fundamental cycle
        int distTo = findDistanceTo(nodeTo,commonAncestor,treeParent);
        int distFrom = findDistanceTo(nodeFrom,commonAncestor,treeParent);
        int[] fundCycle = new int[1+distTo+distFrom];
        fundCycle[0] = edgeId;
        for (int i = 1; i <= distTo; i++) {
            fundCycle[i] = edgeToParent[nodeTo];
            nodeTo = treeParent[nodeTo];
        }
        for (int i = distTo+distFrom; i >= 1+distTo ; i--) {
            fundCycle[i] = -edgeToParent[nodeFrom];
            nodeFrom = treeParent[nodeFrom];
        }
        addCycleToListIfNotPresent(fundCycle,fundCycle.length,fundamentalCycles);
    }

    private static int findDistanceTo(int nodeInTree, int ancestor, int[] treeParent){
        int distance = 0;
        while(nodeInTree!=ancestor){
            distance++;
            nodeInTree = treeParent[nodeInTree];
        }
        return distance;
    }

    private static int findCommonAncestor(int nodeA, int nodeB, int[] treeParent){
        boolean[] visited = new boolean[treeParent.length];
        int curNode = nodeA;
        // tag all nodes above nodeA
        while(true){
            visited[curNode] = true;
            curNode = treeParent[curNode];
            if(curNode==-1){
                break;
            }
        }
        // search nodes above nodeB
        curNode = nodeB;
        while(true){
            if(visited[curNode]){
                return curNode;
            } else {
                curNode = treeParent[curNode];
            }
        }
    }


    public static LinkedList<int[]> generateCyclesOfMaxLength(IndexedInstance indexedInstance, int maxLen){
        // maxLen >= 3
        LinkedList<int[]> result = new LinkedList<>();
        int[] cycle = new int[maxLen];
        boolean[] nodeForbidden = new boolean[indexedInstance.nodes()+1];
        boolean[] edgeForbidden = new boolean[indexedInstance.edges()+1];
        for (int edgeId = 1; edgeId <= indexedInstance.edges(); edgeId++) {
                cycle[0] = edgeId;
                edgeForbidden[Math.abs(edgeId)] = true;
                finishCycle(indexedInstance.edgeIdFrom(edgeId),cycle,1,maxLen,edgeForbidden,nodeForbidden,result, indexedInstance);
                edgeForbidden[Math.abs(edgeId)] = false;
        }
        return result;
    }

    private static void finishCycle(int startNode,int[] cycle, int cycleLength, int maxLen,
                                   boolean[] edgeForbidden, boolean[] nodeForbidden, LinkedList<int[]> result, IndexedInstance indexedInstance){
        int currentNode = indexedInstance.edgeIdTo(cycle[cycleLength-1]);
        /*System.out.println();
        for (int i = 0; i < cycleLength; i++) {
            System.out.print(" -> " + cycle[i]);
        }*/
        if(nodeForbidden[currentNode]) {
            return;
        }
        nodeForbidden[currentNode] = true;
        // for all outgoing edges from currentNode
        for (int i = 0; i < indexedInstance.incidentEdgesToNodeId(currentNode).length; i++) {
            int edgeId = indexedInstance.incidentEdgesToNodeId(currentNode)[i];
            if(edgeForbidden[Math.abs(edgeId)]){
                continue;
            }
            edgeForbidden[Math.abs(edgeId)] = true;
            if(indexedInstance.edgeIdTo(edgeId) == startNode){
                // cycle is closed
                cycle[cycleLength] = edgeId;
                addCycleToListIfNotPresent(cycle,cycleLength+1,result);
            } else if (cycleLength < maxLen-1){
                cycle[cycleLength] = edgeId;
                finishCycle(startNode,cycle,cycleLength+1,maxLen,edgeForbidden,nodeForbidden,result, indexedInstance);
            }
            edgeForbidden[Math.abs(edgeId)] = false;
        }
        nodeForbidden[currentNode] = false;
    }

    private static void addCycleToListIfNotPresent(int[] cycle, int cycleLength, LinkedList<int[]> result){
        int minIndex = 0;
        for (int i = 0; i < cycleLength; i++) {
            if(Math.abs(cycle[minIndex]) > Math.abs(cycle[i])){
                minIndex = i;
            }
        }
        int[] cycleRotated = new int[cycleLength];
        if(cycle[minIndex] > 0) {
            for (int i = 0; i < cycleLength; i++) {
                cycleRotated[i] = cycle[(minIndex + i) % (cycleLength)];
            }
        } else {
            for (int i = 0; i < cycleLength; i++) {
                cycleRotated[i] = -cycle[(2*cycleLength + minIndex - i) % (cycleLength)];
            }
        }
        boolean present = false;
        for (int[] presentCycle:result) {
            if(presentCycle.length == cycleRotated.length){
                boolean same = true;
                for (int i = 0; i < cycleRotated.length; i++) {
                    if(cycleRotated[i] != presentCycle[i]){
                        same = false;
                        break;
                    }
                }
                if(same){
                    present = true;
                    break;
                }
            }
        }
        if(!present){
            result.add(cycleRotated);
        }
    }


}
